@@ -2,7 +2,7 @@
 # Client-to-Client-Protocol (CTCP) support.
 import pydle.protocol
 from pydle.features import rfc1459
-
+from pydle import client
 __all__ = [ 'CTCPSupport' ]
 
 
@@ -15,7 +15,7 @@ class CTCPSupport(rfc1459.RFC1459Support):
 
     ## Callbacks.
 
-    def on_ctcp(self, by, target, what, contents):
+    async def on_ctcp(self, by, target, what, contents):
         """
         Callback called when the user received a CTCP message.
         Client subclasses can override on_ctcp_<type> to be called when receiving a message of that specific CTCP type,
@@ -23,7 +23,7 @@ class CTCPSupport(rfc1459.RFC1459Support):
         """
         pass
 
-    def on_ctcp_reply(self, by, target, what, response):
+    async def on_ctcp_reply(self, by, target, what, response):
         """
         Callback called when the user received a CTCP response.
         Client subclasses can override on_ctcp_<type>_reply to be called when receiving a reply of that specific CTCP type,
@@ -31,34 +31,34 @@ class CTCPSupport(rfc1459.RFC1459Support):
         """
         pass
 
-    def on_ctcp_version(self, by, target, contents):
+    async def on_ctcp_version(self, by, target, contents):
         """ Built-in CTCP version as some networks seem to require it. """
         import pydle
 
         version = '{name} v{ver}'.format(name=pydle.__name__, ver=pydle.__version__)
-        self.ctcp_reply(by, 'VERSION', version)
+        await self.ctcp_reply(by, 'VERSION', version)
 
 
     ## IRC API.
 
-    def ctcp(self, target, query, contents=None):
+    async def ctcp(self, target, query, contents=None):
         """ Send a CTCP request to a target. """
         if self.is_channel(target) and not self.in_channel(target):
             raise client.NotInChannel(target)
 
-        self.message(target, construct_ctcp(query))
+        await self.message(target, construct_ctcp(query, contents))
 
-    def ctcp_reply(self, target, query, response):
+    async def ctcp_reply(self, target, query, response):
         """ Send a CTCP reply to a target. """
         if self.is_channel(target) and not self.in_channel(target):
             raise client.NotInChannel(target)
 
-        self.notice(target, construct_ctcp(query, response))
+        await self.notice(target, construct_ctcp(query, response))
 
 
     ## Handler overrides.
 
-    def on_raw_privmsg(self, message):
+    async def on_raw_privmsg(self, message):
         """ Modify PRIVMSG to redirect CTCP messages. """
         nick, metadata = self._parse_user(message.source)
         target, msg = message.params
@@ -70,29 +70,30 @@ class CTCPSupport(rfc1459.RFC1459Support):
             # Find dedicated handler if it exists.
             attr = 'on_ctcp_' + pydle.protocol.identifierify(type)
             if hasattr(self, attr):
-                getattr(self, attr)(nick, target, contents)
+                await getattr(self, attr)(nick, target, contents)
             # Invoke global handler.
-            self.on_ctcp(nick, target, type, contents)
+            await self.on_ctcp(nick, target, type, contents)
         else:
-            super().on_raw_privmsg(message)
+            await super().on_raw_privmsg(message)
 
-    def on_raw_notice(self, message):
+
+    async def on_raw_notice(self, message):
         """ Modify NOTICE to redirect CTCP messages. """
         nick, metadata = self._parse_user(message.source)
         target, msg = message.params
 
         if is_ctcp(msg):
             self._sync_user(nick, metadata)
-            type, response = parse_ctcp(msg)
+            _type, response = parse_ctcp(msg)
 
             # Find dedicated handler if it exists.
-            attr = 'on_ctcp_' + pydle.protocol.identifierify(type) + '_reply'
+            attr = 'on_ctcp_' + pydle.protocol.identifierify(_type) + '_reply'
             if hasattr(self, attr):
-                getattr(self, attr)(user, target, response)
+                await getattr(self, attr)(nick, target, response)
             # Invoke global handler.
-            self.on_ctcp_reply(user, target, type, response)
+            await self.on_ctcp_reply(nick, target, _type, response)
         else:
-            super().on_raw_notice(message)
+            await super().on_raw_notice(message)
 
 
 ## Helpers.
