@@ -1,12 +1,10 @@
 ## tls.py
 # TLS support.
-import ssl
-
 import pydle.protocol
 from pydle.features import rfc1459
 from .. import connection
 
-__all__ = [ 'TLSSupport' ]
+__all__ = ['TLSSupport']
 
 DEFAULT_TLS_PORT = 6697
 
@@ -27,16 +25,16 @@ class TLSSupport(rfc1459.RFC1459Support):
         self.tls_client_cert_key = tls_client_cert_key
         self.tls_client_cert_password = tls_client_cert_password
 
-    def connect(self, hostname=None, port=None, tls=False, **kwargs):
+    async def connect(self, hostname=None, port=None, tls=False, **kwargs):
         """ Connect to a server, optionally over TLS. See pydle.features.RFC1459Support.connect for misc parameters. """
         if not port:
             if tls:
                 port = DEFAULT_TLS_PORT
             else:
                 port = rfc1459.protocol.DEFAULT_PORT
-        return super().connect(hostname, port, tls=tls, **kwargs)
+        return await super().connect(hostname, port, tls=tls, **kwargs)
 
-    def _connect(self, hostname, port, reconnect=False, password=None, encoding=pydle.protocol.DEFAULT_ENCODING, channels=[], tls=False, tls_verify=False, source_address=None):
+    async def _connect(self, hostname, port, reconnect=False, password=None, encoding=pydle.protocol.DEFAULT_ENCODING, channels=[], tls=False, tls_verify=False, source_address=None):
         """ Connect to IRC server, optionally over TLS. """
         self.password = password
 
@@ -53,43 +51,20 @@ class TLSSupport(rfc1459.RFC1459Support):
             self.encoding = encoding
 
         # Connect.
-        self.connection.connect()
-        # Add handlers.
-        self.connection.on('read', self.on_data)
-        self.connection.on('error', self.on_data_error)
+        await self.connection.connect()
+
 
     ## API.
 
-    def whois(self, nickname):
-        future = super().whois(nickname)
-
-        # Add field that determines if the target user is connected over TLS.
-        if nickname in self._whois_info:
-            self._whois_info[nickname].setdefault('secure', False)
-
-        return future
+    async def whois(self, nickname):
+        info = await super().whois(nickname)
+        info.setdefault('secure', False)
+        return info
 
 
     ## Message callbacks.
 
-    def on_raw_421(self, message):
-        """ Hijack to ignore absence of STARTTLS support. """
-        if message.params[0] == 'STARTTLS':
-            return
-        super().on_raw_421(message)
-
-    def on_raw_451(self, message):
-        """ Hijack to ignore absence of STARTTLS support. """
-        if message.params[0] == 'STARTTLS':
-            return
-        super().on_raw_451(message)
-
-    def on_raw_670(self, message):
-        """ Got the OK from the server to start a TLS connection. Let's roll. """
-        self.connection.tls = True
-        self.connection.setup_tls()
-
-    def on_raw_671(self, message):
+    async def on_raw_671(self, message):
         """ WHOIS: user is connected securely. """
         target, nickname = message.params[:2]
         info = {
@@ -98,7 +73,3 @@ class TLSSupport(rfc1459.RFC1459Support):
 
         if nickname in self._whois_info:
             self._whois_info[nickname].update(info)
-
-    def on_raw_691(self, message):
-        """ Error setting up TLS server-side. """
-        self.logger.error('Server experienced error in setting up TLS, not proceeding with TLS setup: %s', message.params[0])
